# %% Import packages
import os

import strid

import numpy as np
from itertools import groupby

from sklearn.neighbors import LocalOutlierFactor
from sklearn.cluster import KMeans, AgglomerativeClustering
from fcmeans import FCM
from scipy import stats


import matplotlib.pyplot as plt
from plotting import saveTex
from copy import deepcopy
from scipy.cluster.hierarchy import dendrogram, linkage



# %% Define general properties for comparison
    
def duv(u,v=None,prop=None):
    
    if prop is not None:
        if v is not None: v = getattr(v,prop)
        else:             v = getattr(u.closest_mode,prop)
        u = getattr(u,prop)
        
    return abs(u-v)/max(abs(u), abs(v))

def dmodevec(u,v=None,prop=None):
    if prop is not None:
        if v is not None: v = getattr(v,prop)
        else:             v = getattr(u.closest_mode,prop)
        u = getattr(u,prop)
    return np.abs(u.conj().T@v)**2 / (abs(u.conj().T@u)*abs(v.conj().T@v))

    
def modal_transfer_norm_stochastic(m):
    eig = m.eigenvalue
    phi = np.asarray([m.v]).T*eig
    g = np.asarray([m.g]).T*eig
    z = np.exp(1j * m.wd * m.T)
    M = phi@g.T/(z-eig) + phi@g.T/(2*eig)
    _, s, _ = np.linalg.svd(M)
    return np.amax(s)


#%% Redefine Mode class
class Mode(strid.Mode):
    def __init__(self, eigenvalue, eigenvector, fs=None, g=None, order=None, ind=None, T=None):
        super().__init__(eigenvalue, eigenvector)
        self.fs = fs
        self.moc_power = 10
        self.g = g
        self.order = order
        self.ind = ind
        self.T = T
        self.closest_mode = None
        self.closest_mode_above = None
        self.closest_mode_below = None
        self.possibly_physical = True
        self.friend_above = False
        self.friend_below = False
              
    @property
    def modal_transfer_norm_stochastic(self):
        eig = self.eigenvalue
        phi = np.asarray([self.v]).T*eig
        g = np.asarray([self.g]).T*eig
        z = np.exp(1j * self.wd * self.T)
        M = phi@g.T/(z-eig) + phi@g.T/(2*eig)
        _, s, _ = np.linalg.svd(M)
        return np.amax(s)

    @property
    def mtn_s(self):
        return self.modal_transfer_norm_stochastic

    @property
    def ov(self,power=10): #observability vector i MOC criteria
        return np.concatenate([self.eigenvector*(np.exp(self.eigenvalue/self.fs))**i for i in range(self.moc_power)]) #Observability vector
    @classmethod
    def find_modes_from_ss(cls, A, C, G, fs, order=None, T=None):
        """Return modes from the (discrete) system matrices A and C

        This method finds all modes from the discrete state space system
        matrices A and C.

        Arguments
        ---------
        A : 2darray
            Discrete time state space matrix
        C : 2darray
            Output influence matrix
        fs : float
            Sampling rate

        Returns
        -------
        list
            List of modes (Mode objects)
        """
        lr, Q = np.linalg.eig(A)
        u = fs*np.log(lr)
        Phi = C.dot(Q)
        g = np.linalg.solve(Q,G)
        return [cls(ui, q, g=gi, order=order, ind=ind, fs=fs, T=T) for ind, (ui, q, gi) in enumerate(zip(u, Phi.T, g))] 
    
    def __str__(self):
        return f'Mode object with f={self.f} and damping={self.xi}'

    def __repr__(self):
        return f'Mode(f={self.f:.3f}, n={self.order:.0f}, xi={self.xi:.3f})'

#%% Redefine Stabilization diagram
class Stab(strid.StabilizationDiagram):
    def __str__(self): 
        return 'Stabilization diagram with methods to filter and cluster modes'
    def __init__(self, modes, figure_name="Stabilization Diagram", create_figure=True):
        if create_figure is True: super().__init__(figure_name)
        self.path = "stabdiag/autostrid/"

        self.figure_name = figure_name
        self.linewidth = 1.0
        self.line_alpha = 0.5
        
        self.modes = modes #all modes
        self.pp_modes = modes #possibly physical modes
        self.cs_modes = {}  #certainly spurious modes
        for key in self.pp_modes: self.cs_modes[key] = []
        self.clustered_modes = {}
        self.clustered_modes_representative = {}

        
        self.plotsettings = {"cs":         {'marker'    : self.marker,
                                            'alpha'     : 0.5,
                                            'ms'        : self.markersize//2,
                                            'picker'    : True,
                                            'pickradius': self.pickradius},
                             "pp":         {'marker'    : self.marker,
                                            'alpha'     : 1.0,
                                            'ms'        : self.markersize,
                                            'picker'    : True,
                                            'pickradius': self.pickradius},
                             "clusters":   {'marker'    : self.marker,
                                            'alpha'     : 1.0,
                                            'ms'        : self.markersize,
                                            'picker'    : True,
                                            'pickradius': self.pickradius},
                             "pr":         {'marker'    : 'x',
                                            'alpha'     : 1.0,
                                            'ms'        : self.markersize*2,
                                            'picker'    : False,
                                            'pickradius': 0},
                             "true_f":     {'linewidth' : 0.5,
                                            'alpha'     : 1.0,
                                            'picker'    : False,
                                            'pickradius': 0,
                                            'zorder'    : 0}
                             
                             }
        self.plotlabels = {"order": "Order [-]",
                            "f"   : "Frequency [Hz]",
                           "xi"   : r"$\xi$ [-]"}
    #%% Utilitizes  
    def create_figure(self):
        super().__init__(self.figure_name)
    @property
    def nmin(self):
        return min(self.modes.keys())
    @property
    def nmax(self):
        return max(self.modes.keys())
    @property
    def nbeta(self):
        return int(([*self.modes.keys()][-1]-[*self.modes.keys()][0])/(len(self.modes.keys())-1))
    
    def vs_vec(self,vss, mode, othermode):
        vec=[]
        if 0 in vss: vec.append(duv(mode,othermode,prop="order"))           #dorder
        if 1 in vss: vec.append(duv(mode,othermode,prop="eigenvalue"))      #deigenvalue
        if 2 in vss: vec.append(duv(mode,othermode,prop="f"))               #df
        if 3 in vss: vec.append(duv(mode,othermode,prop="xi"))              #dxi
        if 4 in vss: vec.append(1.0 - dmodevec(mode,othermode,prop="eigenvector"))    #nMAC
        if 5 in vss: vec.append(1.0 - mode.mtn_s)                           #nMTNs
        if 6 in vss: vec.append(duv(mode,othermode,prop="mtn_s"))           #dMTNs
        if 7 in vss: vec.append(1.0 - mode.mpc)                             #nMPC
        if 8 in vss: vec.append(mode.mpd)                                   #MPD
        if 12 in vss: vec.append(1.0 - dmodevec(mode,othermode,prop="ov"))             #nMOC
        return vec
        
    def make_mode_spurious(self,mode):
        mode.possibly_physical = False
        order = mode.order
        self.pp_modes[order].remove(mode)
        if self.clustered_modes != {}: self.clustered_modes[mode.ind].remove(mode)
        if mode.f>0: self.cs_modes[order].append(mode)
        if len(self.pp_modes[order])==0: del self.pp_modes[order]
        
    def loop_orders(self, compare="below", direction="up"):
        orders = list(self.pp_modes.keys())
        if compare=="below" and direction =="up":
            sthis, sother = slice(1,None), slice(0,-1)
        elif compare=="above" and direction=="up":
            sthis, sother = slice(0,-1), slice(1,None)
        return zip(orders[sthis],orders[sother])
    
    def get_modes(self,item="pp"):
        if   item=="pp":          modes = [mode for sublist in self.pp_modes.values() for mode in sublist]
        elif item=="cs":          modes = [mode for sublist in self.cs_modes.values() for mode in sublist]
        elif item=="clusters":    modes = [mode for sublist in self.clustered_modes.values() for mode in sublist]
        elif item=="pr":        modes = [mode for mode in self.clustered_modes_representative.values()]
        else: raise ValueError("Not an item")
        return modes


    def find_closest(self,vss=[2,3],compare="below",direction="up",measure = "norm"):
        for order, other_order in self.loop_orders(compare,direction):
            for mode in self.pp_modes[order]:
                distvecs = [self.vs_vec(vss,mode,othermode) for othermode in self.pp_modes[other_order]]
                
                if measure == "norm": ind = np.argmin([np.linalg.norm(vec) for vec in distvecs])
                elif measure == "sum":  ind = np.argmin([sum(vec) for vec in distvecs])
                else: raise ValueError("Not a valid measure")
                
                if compare == "below": mode.closest_mode_below = self.pp_modes[other_order][ind]
                elif compare == "above": mode.closest_mode_above = self.pp_modes[other_order][ind]
                
                if mode.closest_mode is None or np.linalg.norm(self.vs_vec([2,3],mode,mode.closest_mode))<np.linalg.norm(distvecs[ind]):
                    mode.closest_mode = self.pp_modes[other_order][ind]
                    
    def vs_matrix(self,vs=12):
        '''
        Return vs-matrix of all picked representatives
        '''
        representatives = [mode for mode in self.clustered_modes_representative.values()]
        mat = np.diag(np.ones(len(representatives)))        
        for r,representative in enumerate(representatives[:-1]):
            for c,other_representative in enumerate(representatives[r+1:],start=r+1):
                mat[r,c] = self.vs_vec([vs],representative,other_representative)[0]  
        mat += np.triu(mat,k=1).T
        return mat
    #%% Reynders
    def reynders(self):
        #0
        self.find_closest(compare="below",direction="up")
        self.find_closest(compare="above",direction="up")
        #1
        self.cluster_vs(vss=[1,2,3,4,5,6,7,8],method="kmeans")
        self.filter_vh()
        #2
        self.hierarchical_clustering()
        #3
        self.filter_clusters()
        self.get_physical_representative(method="median",prop="xi")


    #%% Yang
    def yang(self):
        #1
        self.find_closest(vss=[2,12],compare="below",direction="up",measure="sum")
        self.find_closest(vss=[2,12],compare="above",direction="up",measure="sum")
        self.cluster_vs(vss=[2,3,12,7],method="cmeans")
        #2
        self.yang_cluster()
        self.remove_outliers_in_clusters()
        self.get_physical_representative(method="median",prop="f")
        self.join_similar_clusters()
        self.get_physical_representative(method="median",prop="f")

    
    #%% 1. Automated clearing/filtering steps
    
    def cluster_vs(self, vss=[1,2,3,4,5,6,7,8],method="kmeans",visualize_transformation=False):
        modes = self.get_modes()
        distances = [self.vs_vec(vss,m,m.closest_mode) for m in modes]
        
        if method == "kmeans":
            kmeans = KMeans(n_clusters=2).fit(distances) 
            spuriousindex = np.argmax(np.sum(np.array(kmeans.cluster_centers_[:]), axis=1))
            index = kmeans.predict(distances)
    
        elif method == "cmeans":
            #Boxcox and normalize
            distances = np.array(distances)+ np.ones_like(distances)*10**-9
            features=[]
            for col in distances.T:
                    feature, _ = stats.boxcox(col)
                    feature = (feature-np.mean(feature))/np.std(feature)
                    features.append(feature)
            distances_transformed = np.column_stack(features)

            #fuzzy cmeans
            fcm = FCM(n_clusters=2)
            fcm.fit(distances_transformed)
            spuriousindex = np.argmax(np.sum(np.array(fcm.centers[:]), axis=1))
            index = fcm.predict(distances_transformed)
            
            #Visualize transformation
            if visualize_transformation is True:
                x,y = [r"d$f$", r"d$\tilde{f}$"],[r"d$\xi$", r"d$\tilde{\xi}$"]
                m = ['o','s','^','d']
                fig, axs = plt.subplots(1,2)
                for p,(f,xi) in enumerate(zip([distances[:,0],distances_transformed[:,0]],[distances[:,1],distances_transformed[:,1]])):
                    ax = axs[p]
                    for i in sorted(np.unique(index),reverse=True): 
                        ax.scatter(f[index==i],xi[index==i],marker=m[i],alpha=0.8)
                    ax.set(xlabel=x[p], ylabel=y[p])
        

        else: raise ValueError("Not a method")
        #Remove spurious modes
        [self.make_mode_spurious(modes[i]) for i, indexi in enumerate(index) if indexi == spuriousindex]

                           
    def filter_vh(self):
        for order, ms in self.pp_modes.items():
            eigenvalues = [mode.eigenvalue for mode in ms]
            for mode in ms:
                if (mode.f > 0
                    and mode.xi > 0 and mode.xi < 0.20
                    and mode.eigenvalue.conjugate() in eigenvalues):
                    continue
                else:
                    self.make_mode_spurious(mode)
    def filter_vss(self,vss,vslims):
        for order, ms in self.pp_modes.items():
            for mode in ms:
                vsvec = self.vs_vec(vss,mode,mode.closest_mode)
                if sum([vs<vslim for vs,vslim in zip(vsvec,vslims)])==len(vsvec):
                    continue
                    
                else:
                    self.make_mode_spurious(mode)
    #%% 2. Clustering steps
       
    def hierarchical_clustering(self, vss=[1, 4],plot=False):
        modes = self.get_modes()
        distances = [sum(self.vs_vec(vss,m,m.closest_mode)) for m in modes]
        distance_threshold = np.mean(distances) + 2*np.std(distances)

        distances = np.array([[sum(self.vs_vec(vss,mode,other_mode)) for mode in modes] for other_mode in modes])
        clustering = AgglomerativeClustering(n_clusters=None, linkage='average', affinity='precomputed', distance_threshold=distance_threshold).fit(distances)

        labels = clustering.labels_
        clustered_modes = {}

        for label in range(max(labels)+1):
            clustered_modes[label] = []
        for mode, label in zip(modes, labels):
            mode.ind = label
            clustered_modes[label].append(mode)
        if plot is True:            
            fig, ax= plt.subplots(1,1)
            from scipy.spatial.distance import squareform

            condensed_dist = squareform(distances)
            linkresult = linkage(condensed_dist,method='average')
 
            # Control number of clusters in the plot + add horizontal line.
            dendrogram(linkresult, color_threshold=distance_threshold, above_threshold_color='grey', ax=ax,no_labels=True,truncate_mode = 'level', p=80 )
            ax.axhline(y=distance_threshold, c='grey', lw=1, linestyle='dashed')
            
            # Show the graph
            ax.set_ylim([0,distance_threshold*1.05])
           
        self.clustered_modes = clustered_modes

    def yang_cluster(self,vss=[2,12],print_progress=False):
        modes = list(filter(lambda mode: mode.eigenvalue.imag>0, self.get_modes())) #All possibly physical modes with positive frequency
        distances = [sum(self.vs_vec(vss,m,m.closest_mode)) for m in modes]
        distance_threshold = np.mean(distances) + 2*np.std(distances)
        if print_progress: 
            print("Nmodes",len(modes))
            print("Mean",np.mean(distances))
            print("STd:",np.std(distances))

        modes = [list(modes_at_order) for order,modes_at_order in groupby(modes,key = lambda mode: mode.order)]
        
        clusters = [[mode] for mode in modes[0]]
        # print(clusters)
        for modes_at_order in modes[1:]:
            for mode in modes_at_order:
                inds = []
                dists = []
                #find distance to close clusters
                for ind,cluster in enumerate(clusters):
                    distance = sum([sum(self.vs_vec(vss,mode,cluster_mode)) for cluster_mode in cluster])/len(cluster)
                    if distance<distance_threshold:
                        inds.append(ind)
                        dists.append(distance)
                #insert mode in closest cluster
                if len(inds)==0: 
                    if print_progress: print("New cluster")
                    clusters.append([mode])
                elif len(inds)==1: 
                    if print_progress: print("Wanted once")
                    clusters[inds[0]].append(mode)
                elif len(inds)>1: 
                    if print_progress: print(inds,dists), print("Possibly added more than once") 
                    clusters[inds[np.argmin(dists)]].append(mode)
                 
        N_T = 0.3* (self.nmax-self.nmin)/self.nbeta  
        cs_clusters = list(filter(lambda cluster: len(cluster)<=N_T, clusters))
        [self.make_mode_spurious(mode) for cluster in cs_clusters for mode in cluster]
        pp_clusters = list(filter(lambda cluster: len(cluster)>N_T, clusters))
    
        self.clustered_modes = dict([[i,c] for i,c in  enumerate(sorted(pp_clusters,key=lambda cluster: np.mean([mode.f for mode in cluster])))])
        for key,cluster in self.clustered_modes.items():
            for mode in cluster: mode.ind = key
            
    #%% 3. Filter clusters        
    def filter_clusters(self):
        clusters = [cluster for cluster in self.clustered_modes.values()]
        lengths = np.array([len(cluster) for cluster in self.clustered_modes.values()])
        longest = max(lengths)
        n = len(np.where(lengths>longest//5)[0])
        lengths = [[length] for length in np.append(lengths,[0]*n)]
        kmeans = KMeans(n_clusters=2).fit(lengths)

        spuriousindex = np.argmin(np.sum(np.array(kmeans.cluster_centers_[:]), axis=1))

        indexs = kmeans.predict(lengths)

        for i, indexi in reversed(list(enumerate(indexs))):
            if indexi == spuriousindex and lengths[i][0]!=0:
                for mode in self.clustered_modes[i]: self.make_mode_spurious(mode)
                del clusters[i]
                
        self.clustered_modes = dict([[i,c] for i,c in  enumerate(sorted(clusters,key=lambda c: np.mean([mode.f for mode in c])))])
        for key,cluster in self.clustered_modes.items(): 
            for mode in cluster: mode.ind = key
        
        
    def remove_outliers_in_clusters(self):      
        for key, cluster in list(self.clustered_modes.items()):
            if len(cluster)<=2: 
                [self.make_mode_spurious(mode) for mode in cluster]
                del self.clustered_modes[key]
            else:
                clf = LocalOutlierFactor(n_neighbors=len(cluster)-1,contamination=0.25)
                ind_f = clf.fit_predict([[m.f] for m in cluster])
                ind_xi = clf.fit_predict([[m.xi] for m in cluster])
                
                for i,(f,xi) in reversed(list(enumerate(zip(ind_f,ind_xi)))): 
                    if f==-1 or xi==-1:
                        self.make_mode_spurious(self.clustered_modes[key][i])
        
    def join_similar_clusters(self):
        keys = list(self.clustered_modes.keys())
        moc_matrix = self.vs_matrix(vs=12)
        adder_dict = {}
        added = []
        for r,row in enumerate(moc_matrix+np.tril(np.ones_like(moc_matrix))):
            adder_dict[r]=[]
            for col,item in enumerate(row):
                if item<0.2 and col not in added and r not in added: #moc>0.8 -> nmoc<0.2
                    adder_dict[r].append(col)
                    added.append(col)
        
        for r,cols in adder_dict.items():       
            if len(cols)>0:
                print("Joining cluster %i and"%(r),cols)
                new_cluster = self.clustered_modes[keys[r]] 
                for col in cols: 
                    for mode in self.clustered_modes[keys[col]]: mode.ind=keys[r]
                    new_cluster += self.clustered_modes[keys[col]]
                    del self.clustered_modes[keys[col]]

                self.clustered_modes[keys[r]] = new_cluster
            
            
    #%% Get representative mode
    def get_physical_representative(self, method="median",prop="f"):
        self.clustered_modes_representative = {}
        for key, modes in self.clustered_modes.items():
            if method == "median":
                props = [getattr(mode,prop) for mode in modes]
                i = np.argsort(props)[len(props)//2]
            elif method == "max":
                props = [getattr(mode,prop) for mode in modes]
                i = np.argsort(props)[-1]
            elif method == "median_mean_damping":                
                modes = sorted(modes,key = lambda m: m.xi)
                l = len(modes)
                i = l//2
                mode = deepcopy(modes[i])
                mode.eigenvector = np.mean(np.array([m.v for m in modes[i-l//6:i+l//6]]),axis=0)
                modes[i]=mode           
            else:
                print("Not a method")  
            self.clustered_modes_representative[key] = modes[i]    
     

    #%% Plot
    def color(self,item, mode=None):
        if item=="pp":
            color = self.stable_color
        elif item=="cs":
            color = "black" # self.unstable_color
        elif item=="clusters":
            color = "C%i"%mode.ind #if mode.ind!=0 else (0.211,0.211,0.211,0.1)
        elif item=="pr":
            color="C%i"%mode.ind
        elif item=="true_f":
            color = "black"
        return color
    
    def plot(self,item=None,x="f",y="order",*arg,**kwargs):   
        if item is None:
            #plot
            self.plot("cs")
            self.plot("clusters")
            self.plot("pr")
            self.quiver("clusters")
        else:
            modes = self.get_modes(item)
            for mode in modes:
                lines = self.axes_plot.plot(getattr(mode,x), getattr(mode,y), color = self.color(item,mode), **{**self.plotsettings[item], **kwargs})
                lines[0].mode = mode
            self.axes_plot.xaxis.set_visible(True)
            self.axes_psd.xaxis.set_visible(False)
            self.axes_plot.set(xlabel=self.plotlabels[x], ylabel=self.plotlabels[y])

    def plot_truef(self,true_fs):
        for f in true_fs:
            self.axes_plot.axvline(f, color=self.color("true_f"),**{**self.plotsettings["true_f"]})
    
    def quiver(self,item,x="f",y="order",*args,**kwargs):
        for mode in self.get_modes(item):
            if type(getattr(mode,"closest_mode_below")) is Mode and mode.closest_mode_below.possibly_physical:
                self.axes_plot.plot([getattr(getattr(mode,"closest_mode_below"),x), getattr(mode,x)], [getattr(getattr(mode,"closest_mode_below"),y), getattr(mode,y)], color= self.color(item,mode),linewidth=1.0,zorder=10) 

    def clear_plot(self):
        self.axes_plot.cla()
    def reopen_plot(self):
        self.figure.show()
        
    #%% Output and tikz
    def createTikz(self,save_name = None, psd=False, compile_tikz=True, open_compiled_tikz=True,width=10,height=None,**kwargs):
        #Delaxes
        if psd is False: [self.figure.delaxes(ax) for ax in self.figure.axes if ax!=self.axes_plot]
        else: self.figure.delaxes(self.axes_table)
        #Give name
        if save_name is None: save_name = "%s"%(self.figure.get_label())
        #Save
        saveTex(save_name,height=height,width=width,**kwargs)
        #Compile    
        if compile_tikz is True: os.system("pdflatex --enable-write18 --extra-mem-top=1000000000 --synctex=1  %s.tex"%save_name)
        #Open
        if open_compiled_tikz is True: os.startfile("%s.pdf"%save_name)