#%% Import packages
import numpy as np
import pandas as pd
from scipy import signal
import strid
from autostrid import Mode,Stab
import matplotlib.pyplot as plt
from copy import deepcopy
from win10toast import ToastNotifier
import os
import pickle
import time
start = time.process_time()

#%% Parameters
toaster = ToastNotifier()

n_br = 60

factor_fs = 1.5
min_order,max_order,step_order = 10,60,1
runs = 200

matrices = {}
methods = ["reynders","yang"]

results ={}
for method in methods:
    results[method]= {"stab":[],"time":[]}
    
#%% Create frame
m,k = 1e3,1e7
sf = strid.utils.ShearFrame(9, m, k)
sf.set_rayleigh_damping_matrix([sf.get_natural_frequency(1), sf.get_natural_frequency(sf.n)], [.05]*2)
results["sf"] = sf

#True properties
true_f = np.array([sf.get_natural_frequency(i)/(2*np.pi) for i in range(1, sf.n+1)])
true_xi = np.array([sf.get_rayleigh_damping_ratio(i) for i in range(1, sf.n+1)])
true_v= np.array([sf.get_mode_shape(i) for i in range(1, sf.n+1)])

#%% Determine the time discretization and period
nyquist_rate = 2*max(true_f)
fs = nyquist_rate*factor_fs
# T = 100/ min(true_f)
T = 600
t = np.arange(0., T, 1/fs)


for n in range(runs):
    #%% Define loads
    ## Unmeasureable: Stochastic loads on all floors (Process noise)
    w = np.random.normal(size=(sf.n, t.size)) * 1e-1
    
    ## Load matrix, f
    F = w.copy()
    
    # Simulate response, accelerations at each floor measured
    y0, _, _ = sf.simulate(t, F)
    
    noise_std = y0.std()
    
    # Add measurement noise to acceleration 
    v = np.random.normal(size=y0.shape)*noise_std
    y = y0 + v 
    
    #%% Create modes
    ssid = strid.CovarianceDrivenStochasticSID(y, fs)
    
    
    modes = {}
    for i, order in enumerate(range(min_order, max_order, step_order)):
        A, C, G, R0 = ssid.perform(order, n_br)
        modes[order] = Mode.find_modes_from_ss(A, C, G, fs, order=order, T=T)
        modes[order] = [mode for mode in modes[order] if np.isnan(mode.f) == False]
    modes = [modes, deepcopy(modes)]
    
    
    #%% Create stab
    stabdiags = {}
    
    for m,method in enumerate(methods):
        methodtime = time.process_time()
        stabdiags[method] = Stab(modes[m], figure_name=method,create_figure=False)
        getattr(stabdiags[method],method)()
        
        #%% Qantify results
        stabdiags[method].set_true_fxiv(true_f,true_xi,true_v)
        stabdiags[method].match_true_fxiv()
        
        results[method]["stab"].append(stabdiags[method])
        results[method]["time"].append(time.process_time() - methodtime)
    
    print("Finished:", n , "At time:",time.process_time())
    


#%% Export results

with open("results.dict","wb") as file:
    modes = pickle.dump(results,file)
    # with open("results.dict","rb") as file:
    #     results = pickle.load(file)
toaster.show_toast("Finshed: %s "%os.path.basename(__file__),"And that's nice")
