#%% Import packages
import numpy as np
import pandas as pd
from scipy import signal
import strid
from autostrid import Mode,Stab
import matplotlib.pyplot as plt
from copy import deepcopy
from win10toast import ToastNotifier
import os
import pickle
import time
start = time.process_time()

#%% Parameters
toaster = ToastNotifier()

n_br = 60
order_min,order_max,step = 5,20,1
orders = range(order_min, order_max, step)

factor_fs = np.arange(1,2.55,0.125)
max_orders = [order_max+i*10 for i in range(len(factor_fs))]
xm,ym = np.meshgrid(np.array(factor_fs),np.array(max_orders),indexing="ij")

matrices = {}
methods = ["reynders","yang"]

matrices['stab'] = np.empty(xm.shape,dtype=object)    
matrices['time'] = np.empty(xm.shape,dtype=float)    


results = dict([["reynders", deepcopy(matrices)],["yang", deepcopy(matrices)],["time",np.empty(xm.shape)],["xm",xm],["ym",ym]])
    
#%% Create frame
m,k = 1e3,1e7
sf = strid.utils.ShearFrame(9, m, k)
sf.set_rayleigh_damping_matrix([sf.get_natural_frequency(1), sf.get_natural_frequency(sf.n)], [.05]*2)
results["sf"] = sf

#True properties
true_f = np.array([sf.get_natural_frequency(i)/(2*np.pi) for i in range(1, sf.n+1)])
true_xi = np.array([sf.get_rayleigh_damping_ratio(i) for i in range(1, sf.n+1)])
true_v= np.array([sf.get_mode_shape(i) for i in range(1, sf.n+1)])

#%% Determine the time discretization and period
nyquist_rate = 2*max(true_f)
factor_max = 10
fs0 = factor_max * max(true_f)
# T = 100/ min(true_f)
T = 600
t0 = np.arange(0., T, 1/fs0)
rs = fs0/min(true_f)

#%% Define loads
## Unmeasureable: Stochastic loads on all floors (Process noise)
w = np.random.normal(size=(sf.n, t0.size)) * 1e-1

## Load matrix, f
F = w.copy()

# Simulate response, accelerations at each floor measured
y0, _, _ = sf.simulate(t0, F)

noise_std = y0.std()

# Add measurement noise to acceleration 
v = np.random.normal(size=y0.shape)*noise_std
y = y0 + v 

steptime = time.process_time()
# fig,ax =plt.subplots(1,1)
for r,f_fs in reversed(list(enumerate(factor_fs))):
    fs = f_fs * nyquist_rate
    t = np.arange(0., T, 1/fs)
    
    yr = y.copy()
    yr = signal.resample(yr,len(t),axis=1)


    for c,max_order in reversed(list(enumerate(max_orders))):
        # plt.close('all')
    
        
        #%% Create modes
        ssid = strid.CovarianceDrivenStochasticSID(yr, fs)
        
        
        modes = {}
        for i, order in enumerate(range(order_min, max_order, step)):
            A, C, G, R0 = ssid.perform(order, n_br)
            modes[order] = Mode.find_modes_from_ss(A, C, G, fs, order=order, T=T)
            modes[order] = [mode for mode in modes[order] if np.isnan(mode.f) == False]
        modes = [modes, deepcopy(modes)]


        #%% Create stab
        stabdiags = {}
        for m,method in enumerate(methods):
            methodtime = time.process_time()
            stabdiags[method] = Stab(modes[m], figure_name=method,create_figure=False)
            getattr(stabdiags[method],method)()
            
            #%% Qantify results
            
            results[method]["stab"][r,c] = stabdiags[method]
            results[method]["time"][r,c] = time.process_time() - methodtime

        results["time"][r,c] = time.process_time() - steptime
        steptime = time.process_time()
        print("Finished:", r,c ,", that is:",f_fs,max_order, "At time:",time.process_time())
        
    
    # f, psd = ssid.psdy(nperseg=2**10*fs/fs0)
    # ax.plot(f,np.trace(psd),alpha=0.2,color="C%i"%(r))


#%% Export results
for key in results.keys():
    name ="results_%s.dict"%key
    with open(name,"wb") as file:
        modes = pickle.dump(results[key],file,protocol=pickle.HIGHEST_PROTOCOL)
toaster.show_toast("Finshed: %s "%os.path.basename(__file__),"And that's nice")
